
// Programme linux de lecture de la temprature pour sonde quipe d'un DS75 sur port srie.
// Ralisation de la sonde et de l'interface I2C : http://col2000.org/tempecom
// (c) Pierre COL - F4EGQ - janvier 2003, v2.0 - col2000@caramail.com

// Licence :
// Je fournis le code source de ce programme pensant qu'il pourra aider les programmeurs,
// mais j'en conserve la paternit ; si vous tes amens  en rutiliser tout ou partie
// et/ou  le modifier, vous devez indiquer le nom de l'auteur et le site (ci-dessus).
// Son utilisation ne doit pas se faire  des fins commerciales et il ne doit pas tre
// publi sous quelle forme que ce soit sans l'autorisation crite de l'auteur.

// - Pour pouvoir lancer ce programme, vous devez tre logu en tant que "root" !
// -  compiler avec la ligne et les options suivantes :
//   gcc -g -O3 -o lireds75 lireds75v2.c

#include <sys/io.h>

#define boolean    int
#define true         1
#define false        0

//-------------------- Prototypes des fonctions : ----------------------------

void            Power_OFF(unsigned short int);
int             Power_ON(unsigned short int);
void            Clear_TxD(unsigned short int);
void            Set_TxD(unsigned short int);
unsigned char   Get_TxD(unsigned short int);
void            Clear_SDA(unsigned short int);
void            Set_SDA(unsigned short int);
unsigned char   Get_SDA(unsigned short int);
void            Clear_SCL(unsigned short int);
void            Set_SCL(unsigned short int);
unsigned char   Get_T_out(unsigned short int);
unsigned char   Get_Ring(unsigned short int);
unsigned char   Get_DCD(unsigned short int);
void            Start(unsigned short int);
void            Stop(unsigned short int);
void            Clock(unsigned short int);
boolean         Read_ACK(unsigned short int);
void            Write_ACK(unsigned short int);
boolean         Write_Byte(unsigned short int, unsigned char);
unsigned char   Read_Byte(unsigned short int, boolean);
boolean         Select_Register(unsigned short int, int, int);
boolean         Write_Configuration(unsigned short int, int, unsigned char);
int             Read_16bitRegister(unsigned short int, int);
void            DisplayHelp(void);
//              Non utilises dans ce programme :
//boolean       Write_16bitRegister(unsigned short int, int, int, unsigned char, unsigned char);
//int           Read_Configuration(unsigned short int, int);

//-------------------- Implmentation des fonctions : ------------------------

void Power_OFF(unsigned short int port)
{
  Clear_SDA(port);
  Clear_SCL(port);
  Clear_TxD(port);
}

int Power_ON(unsigned short int port)
{
  int            erreurPO;
  unsigned char  test0,test1;

  outb(0x00,port+4);
  Clear_TxD(port);
  usleep(100);
  test0=Get_TxD(port);
  Set_TxD(port);
  usleep(100);
  test1=Get_TxD(port);
  if ((test0==0)&&(test1==1))
  {
    Clear_SDA(port);
    usleep(100);
    test0=Get_SDA(port);
    Set_SDA(port);
    usleep(100);
    test1=Get_SDA(port);
    if ((test0==0)&&(test1==1))
    {
      erreurPO=0;
      Set_SCL(port);
      Set_TxD(port);
      usleep(10000);
    }
    else
    {
      erreurPO=-2;    // Interface dbranche.
      Clear_TxD(port);
      Clear_SCL(port);
      Clear_SDA(port);
    }
  } else erreurPO=-1; // Port srie absent.

  return erreurPO;
}

void Clear_TxD(unsigned short int port)
{
  outb(inb(port+3) & 0xBF,port+3);
}

void Set_TxD(unsigned short int port)
{
  outb(inb(port+3) | 0x40,port+3);
}

unsigned char Get_TxD(unsigned short int port)
{
  return ( (inb(port+3)>>6) & 0x01 );
}

void Clear_SDA(unsigned short int port)
{
  outb(inb(port+4) & 0xFE,port+4);
}

void Set_SDA(unsigned short int port)
{
  outb(inb(port+4) | 0x01,port+4);
}

unsigned char Get_SDA(unsigned short int port)
{
  return ( (inb(port+6)>>4) & 0x01 );
}

void Clear_SCL(unsigned short int port)
{
  outb(inb(port+4) & 0xFD,port+4);
}

void Set_SCL(unsigned short int port)
{
  outb(inb(port+4) | 0x02,port+4);
}

unsigned char Get_T_out(unsigned short int port)
{
  return ( (inb(port+6)>>5) & 0x01 );
}

unsigned char Get_Ring(unsigned short int port)
{
  return ( (inb(port+6)>>6) & 0x01 );
}

unsigned char Get_DCD(unsigned short int port)
{
  return ( (inb(port+6)>>7) & 0x01 );
}

void Start(unsigned short int port)
{
  Set_SDA(port);
  Set_SCL(port);
  Clear_SDA(port);
  Clear_SCL(port);
  Set_SDA(port);
}

void Stop(unsigned short int port)
{
  Clear_SDA(port);
  Set_SCL(port);
  Set_SDA(port);
}

void Clock(unsigned short int port)
{
  Clear_SCL(port);
  Set_SCL(port);
  Set_SCL(port);
  Set_SCL(port);
  Clear_SCL(port);
}

boolean Read_ACK(unsigned short int port)
{
  unsigned char bit;

  Clear_SCL(port);
  Set_SDA(port);
  Set_SCL(port);
  Set_SCL(port);
  bit=Get_SDA(port);
  Clear_SCL(port);
  return (bit==0x00);
}

void Write_ACK(unsigned short int port)
{
  Clear_SDA(port);
  Clock(port);
  Set_SDA(port);
}

boolean Write_Byte(unsigned short int port, unsigned char value)
{
  char bit;

  Clear_SCL(port);
  for (bit=7;bit>=0;bit--)
  {
    if ((value & (1<<bit))==0) Clear_SDA(port); else Set_SDA(port);
    Clock(port);
    Set_SDA(port);
  }
  return Read_ACK(port);
}

unsigned char Read_Byte(unsigned short int port, boolean do_ack)
{
  char bit;
  unsigned char value=0x00;

  Set_SDA(port);
  for (bit=7;bit>=0;bit--)
  {
    Set_SCL(port);
    value+=(unsigned char)(Get_SDA(port)<<bit);
    Clear_SCL(port);
  }
  if (do_ack) Clear_SDA(port);
  Clock(port);
  Set_SDA(port);
  return value;
}

boolean Select_Register(unsigned short int port, int chip_num, int register_num)
{
  boolean OK_SR=false;

  if ((chip_num>=0)&&(chip_num<=7)&&(register_num>=0)&&(register_num<=3))
  {
    Start(port);
    OK_SR=(Write_Byte(port,(unsigned char) (0x90+2*chip_num)));
    if (OK_SR) OK_SR=(Write_Byte(port,(unsigned char) register_num));
    Stop(port);
  }
  return OK_SR;
}

boolean Write_Configuration(unsigned short int port, int chip_num, unsigned char value)
{
  boolean OK_WC=false;

  if ((chip_num>=0)&&(chip_num<=7))
  {
    Start(port);
    OK_WC=(Write_Byte(port,(unsigned char) (0x90+2*chip_num)));
    if (OK_WC)
    {
      OK_WC=(Write_Byte(port,0x01));
      if (OK_WC) OK_WC=(Write_Byte(port,value));
    }
    Stop(port);
  }
  return OK_WC;
}

int Read_16bitRegister(unsigned short int port, int chip_num)
{
  unsigned char   MSB,LSB;
  boolean         OK_16bR=false;
  int             value;

  if ((chip_num>=0)&&(chip_num<=7))
  {
    Set_SDA(port);
    if (Get_SDA(port)==0x01) OK_16bR=true;
    Start(port);
    if ((Get_SDA(port)==0x00) && OK_16bR) OK_16bR=true;
    if (OK_16bR) OK_16bR=(Write_Byte(port,(unsigned char) (0x91+2*chip_num)));
    if (OK_16bR)
    {
      MSB=Read_Byte(port,true);
      LSB=Read_Byte(port,false);
    }
    Stop(port);
  }
  if (OK_16bR) value=256*((int)MSB)+(int)LSB; else value=-1;
  return value;
}

// Routines non utilises dans ce programme :

/*
// "register_num" :  2=T(hyst)  /  3=T(os)
boolean Write_16bitRegister(unsigned short int port, int chip_num, int register_num,
               unsigned char MSB, unsigned char LSB)
{
  boolean OK=false;

  if ((chip_num>=0)&&(chip_num<=7)&&(register_num>=2)&&(register_num<=3))
  {
    Start(port);
    OK=(Write_Byte(port,(unsigned char) (0x90+2*chip_num)));
    if (OK)
    {
      OK=(Write_Byte(port,(unsigned char) register_num));
      if (OK)
      {
        OK=(Write_Byte(port,MSB));
        if (OK) OK=(Write_Byte(port,LSB));
      }
    }
    Stop(port);
  }
  return OK;
}

int Read_Configuration(unsigned short int port, int chip_num)
{
  unsigned char   bytevalue;
  int             intvalue;
  boolean            OK=false;

  if ((sonde>=0)&&(sonde<=7))
  {
    Start(port);
    OK=(Write_Byte(port,(unsigned char) (0x90+2*chip_num)));
    if (OK)
    {
      OK=(Write_Byte(port,0x01));
      Start(port);
      if (OK)
      {
        OK=(Write_Byte(port,(unsigned char) (0x91+2*sonde)));
        if (OK) bytevalue=Read_Byte(port,false);
      }
    }
    Stop(port);
  }
  if (OK) intvalue=(int) bytevalue; else intvalue=-1;
  return intvalue;
}
*/

// ------------------------ Module principal du programme : ----------------------------

void DisplayHelp(void)
{
  printf("\n");
  printf("\n Programme de lecture de la temprature avec la sonde  DS75");
  printf("\n Site web : http://col2000.org/tempecom");
  printf("\n (c) Pierre COL - F4EGQ - janvier 2004");
  printf("\n");
  printf("\n Syntaxe :  lireds75  COMx  Sx  [-i]  [-f]  [-m]  [-d]  [-h] ");
  printf("\n");
  printf("\n    avec :  COMx = Choix du port srie accueillant l'interface");
  printf("\n                   valeurs possibles : COM1, COM2, COM3, ou COM4.");
  printf("\n            Sx   = Numro de la sonde (selon les broches A0/A1/A2)");
  printf("\n                   valeurs possibles : S0, S1, S2, S3, S4, S5, S6 ou S7.");
  printf("\n");
  printf("\n options :  -i   = Masquer l'affichage des entres.");
  printf("\n            -f   = Afficher la temprature en degrs Fareneit.");
  printf("\n            -m   = Faire une seule mesure, et quitter.");
  printf("\n            -d   = Affichage discret, supprime tous les messages sauf T.");
  printf("\n            -h   = Afficher cette page d'aide (syntaxe du programme).");
  printf("\n");
  printf("\n Pour excuter le programme, vous devez tre logu en tant que \"root\".");
  printf("\n\n");
}

int main(int argc, char * * param)
{
  int                  sonde,numport,acces_root,erreur,valeur,n;
  boolean              aide,show_inputs,fareneit,mono,discret,testport;
  unsigned short int   portcom;
  float                T;

  sonde=-1;
  numport=-1;
  erreur=0;

  aide=false;
  show_inputs=true;
  fareneit=false;
  mono=false;
  discret=false;

  if (argc>1) for (n=1;n<argc;n++)
  {
    if (strcmp(param[n],"com1")==0) { numport=1; portcom=0x03F8; } // /dev/ttyS0 (COM1), port 0x3f8, irq 4
    if (strcmp(param[n],"COM1")==0) { numport=1; portcom=0x03F8; }
    if (strcmp(param[n],"com2")==0) { numport=2; portcom=0x02F8; } // /dev/ttyS1 (COM2), port 0x2f8, irq 3
    if (strcmp(param[n],"COM2")==0) { numport=2; portcom=0x02F8; }
    if (strcmp(param[n],"com3")==0) { numport=3; portcom=0x03E8; } // /dev/ttyS2 (COM3), port 0x3e8, irq 4
    if (strcmp(param[n],"COM3")==0) { numport=3; portcom=0x03E8; }
    if (strcmp(param[n],"com4")==0) { numport=4; portcom=0x02E8; } // /dev/ttyS3 (COM4), port 0x2e8, irq 3
    if (strcmp(param[n],"COM4")==0) { numport=4; portcom=0x02E8; }
    if ( (strcmp(param[n],"s0")==0) || (strcmp(param[n],"S0")==0) ) sonde=0;
    if ( (strcmp(param[n],"s1")==0) || (strcmp(param[n],"S1")==0) ) sonde=1;
    if ( (strcmp(param[n],"s2")==0) || (strcmp(param[n],"S2")==0) ) sonde=2;
    if ( (strcmp(param[n],"s3")==0) || (strcmp(param[n],"S3")==0) ) sonde=3;
    if ( (strcmp(param[n],"s4")==0) || (strcmp(param[n],"S4")==0) ) sonde=4;
    if ( (strcmp(param[n],"s5")==0) || (strcmp(param[n],"S5")==0) ) sonde=5;
    if ( (strcmp(param[n],"s6")==0) || (strcmp(param[n],"S6")==0) ) sonde=6;
    if ( (strcmp(param[n],"s7")==0) || (strcmp(param[n],"S7")==0) ) sonde=7;
    if ( (strcmp(param[n],"-I")==0) || (strcmp(param[n],"-i")==0) ) show_inputs=false;
    if ( (strcmp(param[n],"-F")==0) || (strcmp(param[n],"-f")==0) ) fareneit=true;
    if ( (strcmp(param[n],"-M")==0) || (strcmp(param[n],"-m")==0) ) mono=true;
    if ( (strcmp(param[n],"-D")==0) || (strcmp(param[n],"-d")==0) ) discret=true;
    if ( (strcmp(param[n],"-h")==0) ||
         (strcmp(param[n],"-?")==0) ||
         (strcmp(param[n],"-h")==0) ||
         (strcmp(param[n],"-H")==0) ||
         (strcmp(param[n],"/h")==0) ||
         (strcmp(param[n],"/h")==0) ||
         (strcmp(param[n],"--help")==0) ||
         (strcmp(param[n],"-help")==0) ) aide=true;
  }

  if (aide==true) { erreur|=0x20; DisplayHelp(); }
  else
  {
    acces_root=iopl(3);
    if ((acces_root==-1) || (sonde==-1) || (numport==-1))
    {
      if (discret==false)
      {
        DisplayHelp();
        if (acces_root==-1)
          { erreur|=0x01; printf(" - Erreur : vous n'tes pas actuellement logu en tant que \"root\" !\n"); }
        if (numport==-1)
          { erreur|=0x02; printf(" - Erreur : vous n'avez pas indiqu le port srie actif !\n"); }
        if (sonde==-1)
          { erreur|=0x04; printf(" - Erreur : vous n'avez pas indiqu le numro de la sonde !\n"); }
        printf("\n");
      }
    }
    else
    {
      if (discret==false)
      {
        printf("\n");
        printf("Configuration : port srie COM%i (adresse $%X) / sonde n%i.",numport,portcom,sonde);
        if (mono==false) printf("\nPour interrompre les mesures, appuyez sur la touche [Ctrl]...");
        printf("\n\n");
      }
      testport=Power_ON(portcom);
      if (testport==0)
      {
        if (Write_Configuration(portcom,sonde,0x60)) // 0x60=01100000b
        {
          if (Select_Register(portcom,sonde,0x00))
          {
            if (discret==false) printf("Initialisation...\n");
            do
            {
              sleep(1);
              valeur=Read_16bitRegister(portcom,sonde);
              if (valeur>-1)
              {
                T=valeur/256.0;
                if (T>127) T-=256;
                if (fareneit==false) printf("T = %3.4fC",T); else printf("T = %3.4fF",(T*1.8)+32);
                if (show_inputs==true) printf(" / Entres : T(out)=%i Ring=%i DCD=%i",
                       (int) Get_T_out(portcom),(int) Get_Ring(portcom),(int) Get_DCD(portcom));
                printf("\n");
                if (mono==true) valeur=-1;
              }
              else { if (discret==false) printf("Sonde ou interface dbranche : "); }
            }
            while ((inb(0x060)!=29)&&(valeur>-1)); // 29 = scan code touche [Ctrl].
            if (discret==false) printf("fin des mesures...\n\n");
          }
          else { if (discret==false) printf("\nErreur de pointeur ???\n\n"); }
        }
        else { if (discret==false) printf("\nSonde n%i absente !\n\n",sonde); }
        Power_OFF(portcom);
      }
      else
      {
        if ((testport==-1)&&(discret==false))
        {
          erreur|=0x08;
          printf("Erreur : le port srie COM%i n'existe pas sur ce PC !\n\n",numport);
        }
        if ((testport==-2)&&(discret==false))
        {
          printf("Erreur : Il n'y a pas d'interface I2C sur le port srie COM%i !\n\n",numport);
          erreur|=0x10;
        }
      }
    }
  }

  return erreur;
}
